"""
    Temporary file for testing purposes.
"""

import models.univariate as unimodels
import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm

from generation.synthetic import ar_process, ma_process


N_EXPERIMENTS = 2
TRUE_COEFS = [0.4, 0.1]
SAMPLES = 500
PROCESS = 'AR'
BIAS = True

PREDICT_WITH_AVG_COEFS = False

coefs = []
if BIAS:
    biases = []

for _ in range(N_EXPERIMENTS):
    if PROCESS == 'AR':
        data = ar_process(TRUE_COEFS, samples=SAMPLES, noise=0)
        ar_model = unimodels.AutoRegressive(p=len(TRUE_COEFS), bias=BIAS)
        ar_model.fit(data)
        if BIAS:
            biases.append(ar_model.coef[0])
        coefs.append(ar_model.coef[1:])
    elif PROCESS == 'MA':
        data = ma_process(TRUE_COEFS, samples=SAMPLES, noise=0)
        ma_model = unimodels.MovingAverage(q=len(TRUE_COEFS), bias=BIAS)
        ma_model.fit(data)
        if BIAS:
            biases.append(ma_model.coef[0])
        coefs.append(ma_model.coef[1:])
    else:
        raise ValueError("Invalid process type.")

coefs = np.array(coefs).squeeze()
if BIAS:
    biases = np.array(biases).squeeze()

for i in range(len(TRUE_COEFS)):
    plt.hist(coefs[:, i], bins=100)

    plt.axvline(x=TRUE_COEFS[i], color='red')
    plt.axvline(x=np.mean(coefs[:, i]), color='green', linestyle='--')

    plt.legend(["True coefficient", "Mean of coefficients"])
    plt.title("Histogram of coefficients")
    plt.xlabel("Coefficient")
    plt.ylabel("Frequency")
    plt.show()

    print("Mean of coefficient estimates: ", np.mean(coefs[:, i]))
    print("Standard deviation: ", np.std(coefs[:, i]))

if BIAS:
    plt.hist(np.array(biases), bins=100)

    plt.axvline(x=TRUE_COEFS[0], color='red')
    plt.axvline(x=np.mean(biases), color='green', linestyle='--')

    plt.legend(["True bias", "Mean of bias estimates"])
    plt.title("Histogram of biases")
    plt.xlabel("Bias")
    plt.ylabel("Frequency")
    plt.show()

    print("Mean of bias estimates: ", np.mean(biases))
    print("Standard deviation: ", np.std(biases))

# Make predictions using average coefficients
if PREDICT_WITH_AVG_COEFS:
    avg_coefs = np.mean(coefs, axis=0)
    if BIAS:
        avg_bias = np.mean(biases)
    else:
        avg_bias = 0

    if PROCESS == 'AR':
        ar_model.coef = np.concatenate([[avg_bias], avg_coefs])
    elif PROCESS == 'MA':
        pass
    else:
        raise ValueError("Invalid process type.")

ar_model.coef = np.concatenate([[0], TRUE_COEFS])

# Make predictions using one step ahead forecasting
predictions = np.zeros(len(data))
predictions[:len(TRUE_COEFS)] = data[:len(TRUE_COEFS)]
for i in range(len(TRUE_COEFS), len(data)):
    if PROCESS == 'AR':
        initial_condition = data[(i - len(TRUE_COEFS)):i]
        predictions[i] = ar_model.forecast(initial_condition)
    elif PROCESS == 'MA':
        pass
    else:
        raise ValueError("Invalid process type.")

plt.plot(data)
plt.plot(predictions, linestyle='--', color='red')
plt.title(PROCESS + f"({len(TRUE_COEFS)}) process")
plt.xlabel("Time")
plt.ylabel("Value")
plt.legend(["Data", "Predictions", "Free run predictions"])
plt.show()

# Calculate mean absolute error
mae = np.mean(np.abs(data - predictions))
print("Mean absolute error: ", mae)

# Plot the residuals, their histogram, acf and qq plot in subplots
residuals = data - predictions

fig, axes = plt.subplots(2, 2)
axes[0, 0].plot(residuals)
axes[0, 0].set_title("Residuals")
axes[0, 1].hist(residuals, bins=50)
axes[0, 1].set_title("Histogram of residuals")
axes[1, 0].acorr(residuals, maxlags=10)
axes[1, 0].set_title("Autocorrelation of residuals")
axes[1, 1].set_title("QQ plot of residuals")
sm.qqplot(residuals, ax=axes[1, 1], line='s')
# set figure size
fig.set_size_inches(10, 8)

plt.show()

# Check normality of residuals
jb_test = sm.stats.stattools.jarque_bera(residuals)
print("Jarque-Bera test: ")
print("Statistic: ", jb_test[0])
print("p-value: ", jb_test[1])
print("Skewness: ", jb_test[2])
print("Kurtosis: ", jb_test[3])

if jb_test[1] < 0.05:
    print("Residuals are not normally distributed.")
else:
    print("Residuals are normally distributed.")



"""
    Module dedicated for functions used in the generation of synthetic time
    series.
"""
import collections.abc

import numpy as np
import pandas as pd

from typing import Union


def ar_process(coefs: list, samples: int = 100, noise: float = 0
               ) -> np.ndarray:
    """
    Generate synthetic data from an Autoregressive (AR) process of a given
    length and known coefficients, with the possibility of adding noise to
    the measurements.

    Args:
        coefs (list): list with coefficients of lagged measurements of
          the series. The order of the AR process will be defined by the number
          of defined coefficients. For example, if coefs = [0.5, 0.3], the
          generated series will be an AR(2) process, where 0.5 is the
          coefficient of the first lagged measurement and 0.3 is the
          coefficient of the second lagged measurement.
        samples (int): number of data points to be generated. Default is 100.
        noise (float): standard deviation of the noise to be added to the
        measurements. Default is 0, which means no noise.

    Returns:
        series: array with the generated AR process.
    """
    # If coefs is not a list, make it one
    if not isinstance(coefs, collections.abc.Sequence):
        coefs = [coefs]
    # The order of the AR process is the number of coefficients
    order = len(coefs)
    coefs = np.array(coefs)

    y = np.zeros(samples)
    # Initial values y[0, 1, .., order]. These can be thought of as the
    # initial conditions of the AR process.
    y[:order] = [np.random.normal() for _ in range(order)]

    for k in range(order, samples):
        # Get previous values of the series, reversed. This is done to
        # match the order of the coefficients.
        prev_samples = y[(k - order):k][::-1]

        y[k] = np.sum(np.array(prev_samples) * coefs) + np.random.normal()

    # Since the noise is intended to emulate measurement noise, it is
    # added to the measurements after the AR process is generated.
    if noise:
        y += np.random.normal(0, noise, samples)

    return np.array(y)


def ma_process(coefs: list, samples: int = 100, noise: float = 0
               ) -> np.ndarray:
    """
    Generate synthetic data from a Moving Average (MA) process of a given
    length and known coefficients, with the possibility of adding noise to
    the measurements.

    Args:
        coefs (list): list with coefficients of lagged measurements of
          the series. The order of the MA process will be defined by the number
          of defined coefficients. For example, if coefs = [0.5, 0.3], the
          generated series will be an MA(2) process, where 0.5 is the
          coefficient of the first lagged measurement and 0.3 is the
          coefficient of the second lagged measurement.
        samples (int): number of data points to be generated. Default is 100.
        noise (float): standard deviation of the noise to be added to the
          measurements. Default is 0, which means no noise.

    Returns:
        series: array with the generated MA process.
    """
    # If coefs is not a list, make it one
    if not isinstance(coefs, collections.abc.Sequence):
        coefs = [coefs]
    # The order of the MA process is the number of coefficients
    order = len(coefs)
    coefs = np.array(coefs)

    y = np.zeros(samples)
    # White noise series of errors that will be used in the MA process
    nu = [np.random.normal() for _ in range(samples)]
    # Initialize the process for k = 0 where the previous values of nu are
    # zero.
    y[0] = nu[0]
    for k in range(1, samples):
        if k < order:
            # Not all past measurements of nu are available yet, so the
            # previous values of the series are calculated differently.
            prev_samples = nu[0:k][::-1]
            y[k] = np.sum(np.array(prev_samples) * coefs[:k]) + nu[k]
        else:
            # Get previous values of the series, reversed. This is done to
            # match the order of the coefficients.
            prev_samples = nu[(k - order):k][::-1]

            y[k] = np.sum(np.array(prev_samples) * coefs) + nu[k]

    if noise:
        y += np.random.normal(0, noise, samples)

    return np.array(y)


def seasonal_component(samples: int = 100, period: int = 10,
                       amplitude: float = 1, noise: float = 0
                       ) -> np.ndarray:
    """
    Generate a seasonal component of a given length, period and amplitude,
    with the possibility of adding noise to the measurements.

    The nature of the seasonal component is a sine wave with the given period
    and amplitude.

    Args:
        samples (int): number of data points to be generated. Default is 100.
        period (int): period of the seasonal component. Default is 10.
        amplitude (float): amplitude of the seasonal component. Default is 1.
        noise (float): standard deviation of the noise to be added to the
        measurements. Default is 0, which means no noise.

    Returns:
        series: array with the generated seasonal component.
    """
    # The seasonal component is a sine wave with the given period and
    # amplitude
    omega = 2 * np.pi / period
    y = amplitude * np.sin(omega * np.arange(samples))

    if noise:
        y += np.random.normal(0, noise, samples)

    return np.array(y)


def trend_component(samples: int = 100, slope: float = 0.1,
                    intercept: float = 0, noise: float = 0) -> np.ndarray:
    """
    Generate a trend component of a given length, slope and intercept, with
    the possibility of adding noise to the measurements.

    Args:
        samples (int): number of data points to be generated. Default is 100.
        slope (float): slope of the trend component. Default is 0.
        intercept (float): intercept of the trend component. Default is 0.
        noise (float): standard deviation of the noise to be added to the
        measurements. Default is 0, which means no noise.

    Returns:
        series: array with the generated trend component.
    """
    y = np.arange(samples) * slope + intercept

    if noise:
        y += np.random.normal(0, noise, samples)

    return np.array(y)


def discontinuous_timeseries(start_timestamp: Union[str, pd.Timestamp],
                             end_timestamp: Union[str, pd.Timestamp],
                             freq: Union[str, pd.Timedelta],
                             num_discontinuities: int,
                             is_categorical: bool = False) -> pd.Series:
    """
    Generate a random discontinuous time series with given start and end
    timestamps and frequency.

    Args:
        start_timestamp (str or Timestamp): start date of the time series.
        end_timestamp (str or Timestamp): end date of the time series.
        freq (str or Timedelta): frequency of the time series.
        num_discontinuities (int): number of discontinuity points to be
          generated.
        is_categorical (bool): if True, the time series will be categorical
          with levels 'A', 'B', 'C' and 'D'. Default is False.

    Returns:
        ts (Series): discontinuous time series with random data.
    """
    # Create a date range with specified frequency
    if isinstance(start_timestamp, str):
        start_timestamp = pd.Timestamp(start_timestamp)
    if isinstance(end_timestamp, str):
        end_timestamp = pd.Timestamp(end_timestamp)
    if isinstance(freq, str):
        freq = pd.Timedelta(freq)
    date_range = pd.date_range(
        start=start_timestamp, end=end_timestamp, freq=freq)
    # Randomly select discontinuity points
    discontinuity_points = np.random.choice(date_range, num_discontinuities,
                                            replace=False)
    # Create the time series with random data
    if is_categorical:
        data = np.random.choice(['A', 'B', 'C', 'D'], len(date_range))
    else:
        data = np.random.rand(len(date_range))
    ts = pd.Series(data, index=date_range)
    # Drop NaN values at the discontinuity points
    for point in discontinuity_points:
        ts.drop(point, inplace=True)
    return ts


def part_static_timeseries(start_timestamp: Union[str, pd.Timestamp],
                           end_timestamp: Union[str, pd.Timestamp],
                           frequency: Union[str, pd.Timedelta],
                           n_samples_static: int,
                           value_static: float = 1.0) -> pd.DataFrame:
    """Generate a time series with a part that is static.

    Args:
        start_timestamp (Union[str, pd.Timestamp]): Start date of the time
          series.
        end_timestamp (Union[str, pd.Timestamp]): End date of the time series.
        n_samples_static (int): Number of samples that will be static.

    Raises:
        ValueError: Error raised if the number of samples is less than the
          number of static samples.

    Returns:
        pd.DataFrame: Time series with a part that is static.
    """
    if isinstance(start_timestamp, str):
        start_timestamp = pd.Timestamp(start_timestamp)
    if isinstance(end_timestamp, str):
        end_timestamp = pd.Timestamp(end_timestamp)
    if isinstance(frequency, str):
        frequency = pd.Timedelta(frequency)
    # Generate a DatetimeIndex with a frequency of 1 minute
    dt_index = pd.date_range(
        start=start_timestamp, end=end_timestamp, freq=frequency)

    # Create 3 random float columns
    col1 = np.random.randn(len(dt_index))
    col2 = np.random.randn(len(dt_index))
    col3 = np.random.randn(len(dt_index))

    if len(dt_index) < n_samples_static:
        raise ValueError('The number of samples must be greater than the '
                         'number of static samples.')
    # Set one of the columns to a constant value for n_samples_static samples
    col1[:n_samples_static] = value_static

    # Create the DataFrame
    df = pd.DataFrame({'col1': col1, 'col2': col2, 'col3': col3},
                      index=dt_index)

    return df



"""
    Module with functions for filtering samples.
"""

import pandas as pd
import numpy as np

from typing import Union

from processing.timestamps import infer_sampling_time

from pandas.tseries.frequencies import to_offset


def get_continuous_patches(
        data: Union[pd.DataFrame, np.ndarray],
        sampling_time: pd.Timedelta = None,
        min_length: Union[str, int] = 0,
        drop_leading: int = 0,
        drop_trailing: int = 0,
        return_num_index: bool = False
) -> list:
    """
    Function to extract the time limits for all continuous patches of data.

    In the context of industrial processes, each continuous patch may be
    associated with a batch, or simply a period of time where the process is
    continuously active between shutdowns.

    This function is often used in conjunction with a function for removivng
    samples in the data that were collected in periods of shutdown.

    Args:
        data (pd.DataFrame): The data to obtain the start and end timestamps.

        sampling_time (pd.Timedelta): Sampling time. If None, it will be
            inferred from the data index. Defaults to None.

        min_length(str, int): The minimum duration to consider a patch as
        valid. If 'str', it must be a string that can be converted to a
        pd.Timedelta. If 'int', it must be an integer representing the number
        of samples. Defaults to 0, which means that all patches will be
        considered.

        drop_leading (int, optional): Number of samples to drop from
            the start of each patch of data. Defaults to 0.

        drop_trailing (int, optional): Number of samples to drop from the end
            of each patch of data. Defaults to 0.

        return_num_index (bool, optional): If True, the function will return
            the row number of the start and end of each patch,
            instead of the timestamp index (if available).
    Returns:
        patches (list): The list of dicts containing the start and end
            of each patch.
    """
    if isinstance(data.index, pd.DatetimeIndex):
        if sampling_time is None:
            sampling_time = infer_sampling_time(data)
        # All time instants where events are happening, with an indicator of
        # whether that instant is the start (first sample) of an event window.
        idx_disc = data.index.to_series().diff().gt(sampling_time).astype(int)
        if isinstance(min_length, str):
            min_length = pd.to_timedelta(min_length)
        elif min_length == 0:
            min_length = pd.to_timedelta('0S')
        else:
            raise ValueError("min_length must be a string if data index is "
                                "a DatetimeIndex.")
    else:
        idx_disc = data.index.to_series().diff().gt(1).astype(int)
        if isinstance(min_length, str):
            raise ValueError("min_length must be an integer if data index is "
                             "not a DatetimeIndex.")

    # Index of timeline array "t" where each event window begins
    event_starts = np.r_[0, np.where(idx_disc != 0)[0]]
    # Index of timeline array "t" where each event window ends
    # Subtract one because the last event ends on the last sample before the
    # next event starts. Add the last sample "t.shape[0]-1" as the end of the
    # last event.
    event_ends = np.r_[np.where(idx_disc != 0)[0] - 1, idx_disc.shape[0] - 1]
    patches_dicts = []

    # Dict containing start and end of all events
    # Remove samples in the head and tail of the series
    event_starts += drop_leading
    event_ends -= drop_trailing
    if isinstance(data.index, pd.DatetimeIndex):
        for start, end in zip(event_starts, event_ends):
            if end >= start:
                start_time = data.index[start]
                end_time = data.index[end]
                if end_time - start_time >= min_length:
                    if return_num_index:
                        patches_dicts.append({"start": start, "end": end})
                    else:
                        patches_dicts.append(
                            {"start": start_time, "end": end_time})
    else:
        for start, end in zip(event_starts, event_ends):
            if end >= start:
                if end - start >= min_length:
                    patches_dicts.append(
                        {"start": start, "end": end})

    return patches_dicts


def rm_stopped_operation(data: pd.DataFrame, rm_events_mask: np.ndarray,
                         rm_interval_start: Union[str, pd.Timedelta] = "0S",
                         rm_interval_stop: Union[str, pd.Timedelta] = "0S",
                         minimum_interval: Union[str, pd.Timedelta] = "0S",
                         return_shutdown_dict: bool = False,
                         ) -> pd.DataFrame:
    """
    Remove all samples in rm_events_mask, plus/minus stop_interval.

    Args:
        data (pd.DataFrame): data to be processed, must have datetime indexing

        rm_events_mask (ndarray): boolean ndarray with length equal to number
                             of rows in data, where rows to be removed are True

        rm_interval_start (Union[str, pd.Timedelta]): time interval to be
                                removed before the events in the mask

        rm_interval_stop (Union[str, pd.Timedelta]): time interval to be
                                removed after the events in the mask

        minimum_interval (string): mininum duration of a stop

        return_shutdown_dict (bool): if True, returns a dictionary with start
                                    and end indices of all events

    Returns:
        pd.DataFrame: data with the rows of rm_events_mask and samples around
            stop_interval removed, in addition to (optionally) shutdown dict.
    """
    # First of all, we convert the time intervals to pd.Timedelta objects
    if isinstance(rm_interval_start, str):
        rm_interval_start = pd.to_timedelta(rm_interval_start)
    if isinstance(rm_interval_stop, str):
        rm_interval_stop = pd.to_timedelta(rm_interval_stop)
    if isinstance(minimum_interval, str):
        minimum_interval = pd.to_timedelta(minimum_interval)

    dataset = data.copy()
    # All time instants where events are happening
    rm_events_idx = dataset[rm_events_mask].index
    # Sampling time
    t_s = infer_sampling_time(dataset)
    offset = to_offset(t_s)
    freq = offset.freqstr

    # All time instants where events are happening (in the index), with an
    # indicator (series value equal to 1) of whether that instant is the start
    # (first sample) of an event window.
    t = rm_events_idx.to_series().diff().gt(t_s).astype(int)
    # Index of TIMELINE ARRAY "t" where each event window begins
    event_starts = np.r_[0, np.where(t == 1)[0]]
    # Index of TIMELINE ARRAY "t" where each event window ends
    # Subtract one because the last event ends on the last sample before the
    # next event starts. Add the last sample "t.shape[0]-1" as the end of the
    # last event.
    event_ends = np.r_[np.where(t == 1)[0]-1, t.shape[0]-1]

    # List of dicts containing start and end of all events
    shutdown_dicts = []
    for start, end in zip(event_starts, event_ends):
        start_time = t.index[start]
        end_time = t.index[end]
        shutdown_dicts.append({"start": start_time, "end": end_time})
    # All indexes to be removed
    stop_idx = np.empty((0), dtype='datetime64[ns]')

    for event in shutdown_dicts:
        stop_interval = event['end'] - event['start']
        if stop_interval >= minimum_interval:
            start = event['start'] - rm_interval_start
            # Ending moment of the window post event
            end = event['end'] + rm_interval_stop
            interval = pd.date_range(start, end, freq=freq)
            stop_idx = np.r_[stop_idx, np.array(interval)]
            stop_idx = np.unique(stop_idx)

    stop_idx = np.intersect1d(stop_idx, dataset.index)
    dataset = dataset.drop(stop_idx)
    if return_shutdown_dict:
        return dataset, shutdown_dicts
    else:
        return dataset


def filter_static_windows(data: pd.DataFrame, columns: list = None,
                          threshold: int = 10,
                          remove_window: bool = False,
                          return_n_removed: bool = True) -> pd.DataFrame:
    """
    Filter windows where there is no variation for 'threshold' consecutive
        samples. In other words, replace static windows for null values or
        remove windows, where a window is considered static if the value of the
        series remains unchanged for at least 'threshold' samples.

    Args:
        data (pd.DataFrame): Dataframe with the values to be analyzed
        columns (list, optional): Columns to check if the windows are static.
            Defaults to [].
        threshold (int, optional): Minimum length of sequence of samples with
            the same value to remove. Defaults to 10.
        remove_window (bool, optional): If True, removes the static windows.
            If False, replace static windows with NA. Defaults to False.
        return_n_removed (bool, optional): If True, returns the amount of
            removed samples for every column. Defaults to True.

    Returns:
        pd.DataFrame: Dataframe without the static windows (or NAs inplace).
        pd.DataFrame: Columns with amount of static windows removed (or NAs
        inplace).
    """
    if not columns:
        columns = data.columns
    column_time_static = {}
    for column in columns:
        time_static = 0
        series = data[column]
        series_diff = series.diff()
        series_diff_0 = np.where(series_diff == 0)[0]
        if list(series_diff_0):
            dis_points_start = [series_diff_0[0]] + list(series_diff_0[
                np.where(np.diff(series_diff_0,
                                 prepend=series_diff_0[0]) > 1)[0]])
            dis_points_end = (list(series_diff_0[np.where(np.diff(
                series_diff_0, prepend=series_diff_0[0]) > 1)[0] - 1]) +
                [series_diff_0[-1]])

            # Reverse the lists to drop correctly the samples (if we drop the
            # first sample, the index of the second sample is not coherent
            # anymore)
            dis_points_start.reverse()
            dis_points_end.reverse()

            for start, end in zip(dis_points_start, dis_points_end):
                if (end - start) >= threshold:
                    # If remove is True, remove the window
                    if remove_window:
                        # * Drop including the 'end' sample
                        data = data.drop(data[start:end+1].index)
                    else:
                        # * Replace the window with null values
                        data.loc[start:end, column] = np.nan
                    time_static += (end - start)

        column_time_static[column] = time_static
        time_static_pd = pd.DataFrame.from_dict(column_time_static,
                                                orient='index',
                                                columns=['time_static'])
        time_static_pd.sort_values(by='time_static', ascending=False,
                                   inplace=True)

    if return_n_removed:
        return data, time_static_pd
    else:
        return data


def remove_static_columns(df: pd.DataFrame, min_std_cv: float = 0.01,
                          columns: list = None,
                          return_removed: bool = True):
    """
    Removes columns for which the coefficient of variation is below a selected
        threshold.
    
        If the mean of the column is 0, the standard deviation is used instead
        of the coefficient of variation.

    Args:
        df (pd.DataFrame): Process dataframe
        min_std_cv (float, optional): Minimum variation coefficient.
            Defaults to 0.01.
        columns (list, optional): List of columns to check. If None, all
            columns are checked. Defaults to None.
        return_removed (bool, optional): If True, returns the removed columns
            and their coefficient of variation. Defaults to True.

    Returns:
        pd.DataFrame: The dataframe without the columns that don't change
        pd.DataFrame: Dataframe with the columns that don't change and their
            coefficient of variation
    """
    if columns is None:
        columns = df.columns
    column_coef_var = {}
    for col in columns:
        col_mean = df[col].mean()
        col_std = df[col].std()
        if col_mean != 0:
            col_cv = col_std / col_mean
            # ! If the column's standard deviation is 0, it is removed. It's
            # ! not possible to calculate the coefficient of variation in this
            # ! case, so we have to drop the column.
            if np.abs(col_cv) <= min_std_cv:
                df = df.drop(col, axis=1)
                column_coef_var[col] = col_cv
        else:
            if col_std == 0:
                df = df.drop(col, axis=1)
                column_coef_var[col] = 0
            else:
                if np.abs(col_std) <= min_std_cv:
                    df = df.drop(col, axis=1)
                    column_coef_var[col] = col_cv

    coef_var_pd = pd.DataFrame.from_dict(column_coef_var, orient='index',
                                         columns=['coef_var'])
    if return_removed:
        return df, coef_var_pd
    else:
        return df


def format_start(df: pd.DataFrame, s: int = 0, m: int = 0,
                 h: int = 0):
    """
    Function to remove (if necessary) the first rows of the data in
        order to have the first row in a specific format.

    If the formatting is not possible, the function returns the original
        dataframe.

    Args:

        df (pd.DataFrame): df whose index is in datetime format.
        s (int): initial second, with all windows starting with this second.
        m (int): initial minute, with all windows starting with this minute.
        h (int): if 0, initial hours are even; if 1, initial hours are odd.

    Returns:

        pd.DataFrame: with the specified format.

    """
    # hour = 0 -> even hours
    # hour = 1 -> odd hours
    matches = np.array([True] * len(df))
    if s != -1:
        matches = matches & (df.index.second == s)
    if m != -1:
        matches = matches & (df.index.minute == m)
    if h != -1:
        matches = matches & (df.index.hour % 2 == h)

    if matches.any():
        first_match = df.index[matches].min()
        df = df.loc[first_match:]
    return df




"""
    Module with functions for time series feature engineering.
"""
import pandas as pd


def counts_ratio_per_patch(timeseries: pd.Series, patches_dicts: list,
                           column: str,
                           return_timestamps: bool = False) -> pd.DataFrame:
    """
    Calculates the ratio between counts of all values in a patch of data
        Used for series with discrete inputs for data aggregation.

    This function expects a list of dictionaries with the start and end of
        every patch, in the same format as the output of the function
        `industryts.processing.filtering.get_patches_dicts`.

    Each item in the patches_dicts list is a dictionary with the following
        structure:
        {
            'start': start_timestamp,
            'end': end_timestamp
        }

    The function returns a dataframe in which each column stores the ratio
        of a value in the column parameter in every patch. The column names
        follow the convention "category_originalColName".

    Args:
        timeseries (pd.Series): The time series to be processed

        patches_dicts (list): The list containing the start and end of
            every data patch

        column (str): The column name to be used in the ratio calculation

        return_timestamps (bool): If True, the function will return the
            timestamps of the start and end of every patch. Defaults to False.

    Returns:
        df_column_values (pd.DataFrame): A dataframe contaning the ratio of
            column in every data window.
    """

    # Possible values for that column in any patch
    values = list(timeseries.unique())

    # The final list with every ratio value for every patch
    column_values = []

    for event in patches_dicts:
        # Starting moment of the window
        start = event['start']
        # Ending moment of the window
        end = event['end']

        # Create a key for every possible value in that patch
        patch_values = {value: 0 for value in values}

        # Count of each value in that patch
        values_event = timeseries.loc[start:end].value_counts().to_dict()
        size_event = timeseries.loc[start:end].shape[0]

        # * Since the patch does not necessarily has all the values, we use
        # * its value counts to update the patches values
        for key, _ in values_event.items():
            patch_values[key] = values_event[key]/size_event

        if return_timestamps:
            patch_values["start"] = start
            patch_values["end"] = end

        column_values.append(patch_values)

    df_column_values = pd.DataFrame(column_values)

    for df_column in df_column_values.columns:
        df_column_values = df_column_values.rename(
            {
                df_column: f'{column}_{df_column}'
            },
            axis='columns')

    return df_column_values



"""
    Module with functions related to timestamp indices.
"""

import pandas as pd


def infer_sampling_time(data: pd.DataFrame) -> float:
    """
    Infers sampling time from datetime index.
    If the index is not a datetime index, it will raise an error.
    Returns the sampling time as a pd.Timedelta object.

    The algorithm works by taking the first 10 samples and inferring
    the frequency from them. If the frequency is not inferred, it will
    take the next 10 samples and try again. This is repeated until the
    frequency is inferred or all samples are exhausted.

    Args:
        data (pd.DataFrame): Data frame with datetime index to infer
          the sampling time

    Returns:
        t_s (pd.Timedelta): The inferred sampling time
    """
    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("Data index must be a datetime index.")

    freq = pd.infer_freq(data.sort_index().index[:10])
    sample = 1
    while freq is None:
        freq = pd.infer_freq(data.sort_index().index[sample: 10 + sample])
        sample += 1
        if sample > data.shape[0] - 10:
            raise ValueError(
                "Could not infer sampling time. Exhausted all samples.")
    # If the frequency is 1S, this function should return 1S instead of S for
    # consistency.
    if len(str(freq)) == 1:
        freq = f"1{freq}"
    t_s = pd.to_timedelta(freq)

    return t_s



"""
    Test the functions in timefunctions.py
"""
import generation.synthetic as syn
import processing.timestamps as tf
import processing.filtering as flt
import pandas as pd
import processing.featureengineering as fe


"""The expected flow of work is the following:
1. Generate synthetic data without discontinuities
2. Remove some data points to simulate discontinuities (e.g. shutdowns).
In real life we have in fact shutdowns.
3. Get the continuous patches of the time series
4. Compute the ratio of counts per patch in case there are categories
"""

start_date = "2023-01-01"
end_date = "2023-12-31"
frequency = "1D"

# Test function get_continuous_patches ----------------------------------------
n_discontinuities = 40
timeseries = syn.discontinuous_timeseries(start_date, end_date,
                                          frequency, n_discontinuities)
patches = flt.get_continuous_patches(timeseries, frequency)

# Test function counts_ratio_per_patch ----------------------------------------
n_discontinuities = 40
timeseries = syn.discontinuous_timeseries(start_date, end_date,
                                          frequency, n_discontinuities, True)
patches = flt.get_continuous_patches(timeseries, frequency)
column_values = fe.counts_ratio_per_patch(timeseries, patches, 'categories')

# Test function rm_stopped_operation ------------------------------------------
# No discontinuities
n_discontinuities = 0
timeseries_a = syn.discontinuous_timeseries(start_date, end_date,
                                            frequency, n_discontinuities)
timeseries_b = syn.discontinuous_timeseries(start_date, end_date,
                                            frequency, n_discontinuities)
dataframe = pd.DataFrame({'a': timeseries_a, 'b': timeseries_b})
print("=====================================================")
print("Test rm_stopped_operation")
print("Original shape", dataframe.shape)
print("Rows to be removed", dataframe[dataframe['a'] >= 0.8].shape)
# Suppose that we want to remove the batches where a >= 0.8
mask = dataframe['a'] >= 0.8
# Removing only the rows where a >= 0.8
dataframe_0_0 = flt.rm_stopped_operation(dataframe, mask,
                                        return_shutdown_dict=False)
print("Shape after removing only stopped operation", dataframe_0_0.shape)

# Test remove 2D before and after
dataframe_2_2 = flt.rm_stopped_operation(dataframe, mask, "2D", "2D",
                                        return_shutdown_dict=False)
print("Shape after removing 2 days before and after", dataframe_2_2.shape)
patches = flt.get_continuous_patches(timeseries, frequency)
print("Number of patches", len(patches))
print("=====================================================")
# Test function filter_static_windows =========================================
n_samples_static = 100
frequency = "1 minute"
print("Test filter_static_windows")
# Replace the static part with NaNs
timeseries_static = syn.part_static_timeseries(
    start_date, end_date, frequency, n_samples_static)
print("NAs in the static part", timeseries_static.isna().sum())
timeseries_static, removed_df = flt.filter_static_windows(
    timeseries_static, threshold=10)
print("Shape after filtering static part", timeseries_static.shape)
print("NAs after filtering the static part", timeseries_static.isna().sum())
print("Removed dataframe", removed_df)

# Removing in fact the static part
timeseries_static = syn.part_static_timeseries(
    start_date, end_date, frequency, n_samples_static)
print("Original shape", timeseries_static.shape)
timeseries_static, removed_df = flt.filter_static_windows(
    timeseries_static, threshold=10, remove_window=True)
print("Shape after removing static part", timeseries_static.shape)
print("Removed dataframe", removed_df)
print("=====================================================")
# Test function remove_static_columns and remove_almost_static_columns ========
start_date = "2023-01-01"
end_date = "2023-12-31"
frequency = "1D"
n_samples_static = 365
print("Test remove_static_columns and remove_almost_static_columns")
timeseries_static = syn.part_static_timeseries(
    start_date, end_date, frequency, n_samples_static, 1000)
print("Original shape", timeseries_static.shape)
timeseries_static, removed_df = flt.remove_static_columns(timeseries_static)
print("Shape after removing static columns", timeseries_static.shape)
print("Removed dataframe", removed_df)

n_samples_static = 330
timeseries_static = syn.part_static_timeseries(
    start_date, end_date, frequency, n_samples_static, 1000)
print("Original shape", timeseries_static.shape)
timeseries_static = flt.remove_almost_static_columns(timeseries_static)
print("Shape after removing almost static columns", timeseries_static.shape)
print("=====================================================")
# Test function format_start ==================================================
start_date = "2023-01-01"
end_date = "2023-01-31"
frequency = "60S"
print("Test format_start")
n_discontinuities = 100
timeseries_a = syn.discontinuous_timeseries(start_date, end_date,
                                            frequency, n_discontinuities)
patches = flt.get_continuous_patches(timeseries_a, frequency)

# For every patch, we need that it starts at minute 1 at a odd hour
patch = patches[0]
print("Patch", patch)
timeseries_b = timeseries_a[patch["start"]:patch["end"]]
print("Start before format_start", timeseries_b.index[0])
timeseries_c = flt.format_start(timeseries_b, 0, 1, 1)
print("Start after format_start", timeseries_c.index[0])
print("=====================================================")



"""
    Module with classes for univariate models.
"""
import abc

import numpy as np
import pandas as pd

from typing import Union


class UnivariateModel(metaclass=abc.ABCMeta):
    """
    Base class for univariate models.

    Args:
        order (int): Order of the model.
        family (str): Family of the model. Defaults to "AR".

    Attributes:
        order: Order of the model.
        family: Family of the model.
    """

    def __init__(self, order: int, family: str):
        super(UnivariateModel, self).__init__()
        self.__order = order
        self.__family = family

    @abc.abstractmethod
    def fit(self, data: Union[pd.DataFrame, np.ndarray], **kwargs):
        """
        Fit the model to the data.
        """
        pass

    @abc.abstractmethod
    def forecast(self, initial_conditions: Union[pd.DataFrame, np.ndarray],
                 horizon: int = 1) -> np.ndarray:
        """
        Simulate the model forward in time.

        Args:
            initial_conditions (ArrayLike): Initial conditions for the model.
            horizon (int): Number of steps to forecast. Defaults to 1.

        Returns:
            forecast (ndarray): Forecasted values.
        """
        pass

    def _fix_dim_type(self, data: Union[pd.DataFrame, np.ndarray]
                      ) -> np.ndarray:
        if isinstance(data, pd.DataFrame):
            data = data.values
        # If data is 1D, make it 2D
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        return data

    def _prepare_regressors(self, data: np.ndarray,
                            inference: bool = False) -> np.ndarray:
        """
        Prepare the data for prediction.

        Args:
            data (ArrayLike): Data to be predicted.
            inference (bool): Whether the data is being used for inference.

        Returns:
            data (ndarray): Prepared data.
        """
        if inference:
            regressors = data[-self.__order:].reshape(1, -1)
        else:
            regressors = np.hstack(
                [data[i:-self.__order + i] for i in range(self.__order)])

        # Reverse the regressors to match the order of the coefficients
        regressors = regressors[:, ::-1]

        if self._bias:
            regressors = np.hstack([np.ones((regressors.shape[0], 1)),
                                    regressors])
        # *** The order of the returned columns is as follows
        # [bias, y[k-1], y[k-2], ..., y[k-order]]
        return regressors

    def __call__(
            self,
            data: Union[pd.DataFrame, np.ndarray],
            **kwargs) -> np.ndarray:
        self.predict(data, **kwargs)

    def __str__(self) -> str:
        return self.__family + " model of order " + str(self.__order)


class AutoRegressive(UnivariateModel):
    """
    Class for purely autoregressive models of order p.

    Args:
        p (int): Order of the model. Defaults to 1.
        bias (bool): Whether to include a bias term in the model.
            Defaults to True.

    Attributes:
        p: Order of the model.
        coef: Coefficients of the model.
    """

    def __init__(self, p: int = 1, bias: bool = True):
        if p < 1:
            raise ValueError("The order of the model must be greater than 0.")
        if not isinstance(p, int):
            raise TypeError("The order of the model must be an integer.")
        if not isinstance(bias, bool):
            raise TypeError("The bias must be a boolean value.")
        super(AutoRegressive, self).__init__(order=p, family='Autoregressive')
        self.p = p
        self.coef = None
        self._bias = bias

    def fit(self, data: Union[pd.DataFrame, np.ndarray]):
        """
        Fit the model to the data.
        """
        if self.p > data.shape[0]:
            raise ValueError("The order of the model must be less than or "
                             "equal to the number of observations.")

        data = self._fix_dim_type(data)

        regressors = self._prepare_regressors(data)
        targets = data[self.p:]

        self.coef = np.linalg.lstsq(regressors, targets, rcond=None)[0]

    def forecast(self, initial_condition: Union[pd.DataFrame, np.ndarray],
                 horizon: int = 1):
        """
        Simulate the model forward in time.

        Args:
            initial_conditions (ArrayLike): Initial conditions for the model.
            horizon (int): Number of steps to forecast. Defaults to 1.

        Returns:
            forecast (ndarray): Forecasted values.
        """
        if self.coef is None:
            raise ValueError("The model must be fitted before forecasting.")
        if not isinstance(horizon, int):
            raise TypeError("The horizon must be an integer.")
        if horizon < 1:
            raise ValueError("The horizon must be greater than 0.")
        if initial_condition.shape[0] < self.p:
            raise ValueError("The initial conditions must have at least as "
                             "many observations as the order of the model.")

        regressors = self._prepare_regressors(initial_condition,
                                              inference=True)
        forecast = np.zeros((self.p + horizon))
        forecast[:self.p] = initial_condition[-self.p:]

        for i in range(self.p, self.p + horizon):
            forecast[i] = regressors @ self.coef
            regressors = self._prepare_regressors(forecast[-self.p:],
                                                  inference=True)
        return forecast[-horizon:]


class MovingAverage(UnivariateModel):
    """
    Class for purely moving average models of order q.

    Args:
        q (int): Order of the model. Defaults to 1.
        bias (bool): Whether to include a bias term in the model.
            Defaults to True.

    Attributes:
        q: Order of the model.
        coef: Coefficients of the model.
    """

    def __init__(self, q: int = 1, bias: bool = True):
        super(MovingAverage, self).__init__(order=q, family='Moving Average')
        self.q = q
        self.coef = None
        self._bias = bias

    def fit(self, data: Union[pd.DataFrame, np.ndarray],
            n_iterations: int = 100) -> None:
        """
        Fit the model to the data.
        """
        data = self._fix_dim_type(data)
        targets = data[self.q:]

        # Initialize the residuals
        residuals = data
        regressors = self._prepare_regressors(residuals)

        self.coef = [0 for _ in range(self.q)]

        for _ in range(n_iterations):
            # Calculate the coefficients
            self.coef = np.linalg.lstsq(regressors, targets, rcond=None)[0]
            # Calculate the residuals
            residuals = targets - regressors @ self.coef
            # Add q zeros to the beginning of the residuals to match the
            # dimensions of the regressors
            residuals = np.vstack([np.zeros((self.q, residuals.shape[1])),
                                   residuals])
            # Update the regressors
            regressors = self._prepare_regressors(residuals)
